# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..registry import check
from ..error import Error


# Module API

@check('missing-header')
def missing_header(cells, sample):
    errors = []

    for cell in copy(cells):

        # Skip if header in cell
        if cell.get('header') is not None:
            continue

        # Add error
        error = Error('missing-header', cell)
        errors.append(error)

        # Remove cell
        cells.remove(cell)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from simpleeval import simple_eval
from ...registry import check
from ...error import Error


# Module API

@check('custom-constraint', type='custom', context='body')
class CustomConstraint(object):

    # Public

    def __init__(self, constraint, **options):
        self.__constraint = constraint

    def check_row(self, cells):
        # Prepare names
        names = {}
        for cell in cells:
            if None not in [cell.get('header'), cell.get('value')]:
                try:
                    names[cell['header']] = float(cell['value'])
                except ValueError:
                    pass

        # Check constraint
        try:
            # This call should be considered as a safe expression evaluation
            # https://github.com/danthedeckie/simpleeval
            assert simple_eval(self.__constraint, names=names)
        except Exception:
            row_number = cells[0]['row-number']
            message = 'Custom constraint "{constraint}" fails for row {row_number}'
            message_substitutions = {
                'constraint': self.__constraint,
            }
            error = Error(
                'custom-constraint',
                row_number=row_number,
                message=message,
                message_substitutions=message_substitutions
            )
            return [error]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import json
from ..registry import check
from ..error import Error


# Module API

@check('duplicate-row')
class DuplicateRow(object):

    # Public

    def __init__(self, **options):
        self.__row_index = {}

    def check_row(self, cells):
        errors = []

        # Get pointer
        try:
            pointer = hash(json.dumps(list(cell.get('value') for cell in cells)))
            references = self.__row_index.setdefault(pointer, [])
        except TypeError:
            pointer = None

        # Found pointer
        if pointer:

            row_number = cells[0].get('row-number')

            # Add error
            if references:
                message_substitutions = {
                    'row_numbers': ', '.join(map(str, references)),
                }
                error = Error(
                    'duplicate-row',
                    row_number=row_number,
                    message_substitutions=message_substitutions
                )
                errors.append(error)

            # Clear cells
            if references:
                del cells[:]

            # Update references
            references.append(row_number)

        return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from .constraints_checks import create_check_constraint


# Module API

@check('minimum-length-constraint')
def minimum_length_constraint(cells):
    check_constraint = create_check_constraint('minimum-length-constraint', 'minLength')
    return check_constraint(cells)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import re
from copy import copy
from ..registry import check
from ..error import Error


# Module API

@check('non-matching-header')
class NonMatchingHeader(object):

    # Public

    def __init__(self, order_fields=False, **options):
        self.__order_fields = order_fields

    def check_headers(self, cells, sample=None):
        if self.__order_fields:
            return _check_with_ordering(cells)
        else:
            return _check_without_ordering(cells)


# Internal


def _check_with_ordering(cells):
    # Update fields order to maximally match headers order
    headers = [cell.get('header') for cell in cells]
    for index, header in enumerate(headers):
        if header is None:
            continue
        field_name = _get_field_name(cells[index])
        # Cell header and field_name are different
        if _slugify(header) != _slugify(field_name):
            # If there is the match in next cells - swap fields
            for cell in cells[index+1:]:
                # We've found field matching given header
                if _slugify(header) == _slugify(_get_field_name(cell)):
                    _swap_fields(cells[index], cell)
                    break
                if field_name:
                    # Given field matches some header also swap
                    if _slugify(field_name) == _slugify(cell.get('header')):
                        _swap_fields(cells[index], cell)

    # Run check with no field ordering
    return _check_without_ordering(cells)


def _check_without_ordering(cells):
    errors = []

    for cell in copy(cells):
        if cell.get('field') is not None:
            header = cell.get('header')
            if header != cell['field'].name:
                # Add error
                message_substitutions = {
                    'field_name': '"{}"'.format(cell['field'].name),
                }
                error = Error(
                    'non-matching-header',
                    cell,
                    message_substitutions=message_substitutions
                )
                errors.append(error)
                if _slugify(header) != _slugify(cell['field'].name):
                    # Remove cell
                    cells.remove(cell)

    return errors


def _get_field_name(cell):
    if cell.get('field'):
        return cell['field'].name


def _slugify(string):
    if string is None:
        return None
    string = re.sub(r'[\W_]+', '', string)
    string = string.lower()
    return string


def _swap_fields(cell1, cell2):
    field1 = cell1['field']
    cell1['field'] = cell2['field']
    cell2['field'] = field1



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from ..error import Error


# Module API

@check('blank-header')
def blank_header(cells, sample=None):
    errors = []

    for cell in cells:

        # Skip if cell have non blank header
        if cell.get('header', True):
            continue

        # Add error
        error = Error('blank-header', cell)
        errors.append(error)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from ..error import Error


# Module API

@check('duplicate-header')
def duplicate_header(cells, sample=None):
    errors = []

    rindex = {}
    cell_by_column_number = {}
    for cell in cells:

        # Skip if not header
        if 'header' not in cell:
            continue

        header_indexes = rindex.get(cell['header'], set())
        header_indexes.add(cell['column-number'])

        rindex[cell['header']] = header_indexes
        cell_by_column_number[cell['column-number']] = cell

    for header_value, header_indexes in rindex.items():
        if len(header_indexes) == 1:
            continue

        header_indexes_list = sorted(header_indexes)
        first_header_index, other_header_indexes = header_indexes_list[0], header_indexes_list[1:]  # noqa
        for other_header_index in other_header_indexes:
            duplicates = header_indexes - {other_header_index}
            message_substitutions = {
                'original': str(first_header_index),
                'column_numbers': ', '.join(map(str, duplicates)),
            }
            cell = cell_by_column_number[other_header_index]
            error = Error(
                'duplicate-header',
                cell,
                message_substitutions=message_substitutions
            )
            errors.append(error)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


# Module API

from .inspector import Inspector
from .registry import preset, check
from .validate import validate, init_datapackage
from .spec import spec
from .error import Error
from . import exceptions

# Version

import io
import os
__version__ = io.open(
    os.path.join(os.path.dirname(__file__), 'VERSION'),
    encoding='utf-8').read().strip()

# Register

import importlib
from . import config
for module in config.PRESETS:
    importlib.import_module(module)
for module in config.CHECKS:
    importlib.import_module(module)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
from setuptools import setup, find_packages


# Helpers
def read(*paths):
    """Read a text file."""
    basedir = os.path.dirname(__file__)
    fullpath = os.path.join(basedir, *paths)
    contents = io.open(fullpath, encoding='utf-8').read().strip()
    return contents


# Prepare
PACKAGE = 'goodtables'
NAME = PACKAGE.replace('_', '-')
INSTALL_REQUIRES = [
    'six>=1.9',
    'click>=6.6',
    'click-default-group',
    'requests>=2.10',
    'simpleeval>=0.9',
    'statistics>=1.0',
    'tabulator>=1.20',
    'tableschema>=1.0.3',
    'datapackage>=1.2',
]
INSTALL_FORMAT_ODS_REQUIRES = [
    'ezodf>=0.3',
    'lxml>=3.0',
]
TESTS_REQUIRE = [
    'tox',
]
README = read('README.md')
VERSION = read(PACKAGE, 'VERSION')
PACKAGES = find_packages(exclude=['examples', 'tests'])


# Run
setup(
    name=NAME,
    version=VERSION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={
        'develop': TESTS_REQUIRE,
        'ods': INSTALL_FORMAT_ODS_REQUIRES,
    },
    entry_points={
        'console_scripts': [
            'goodtables = goodtables.cli:cli',
        ]
    },
    zip_safe=False,
    long_description=README,
    long_description_content_type='text/markdown',
    description='Goodtables is a framework to inspect tabular data.',
    author='Open Knowledge International',
    author_email='info@okfn.org',
    url='https://github.com/frictionlessdata/goodtables',
    license='MIT',
    keywords=[
        'data validation',
        'frictionless data',
        'open data',
        'json schema',
        'json table schema',
        'data package',
        'tabular data package',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from tabulator import Stream
from functools import partial
from datapackage import Package, exceptions
from ..registry import preset


# Module API

@preset('datapackage')
def datapackage(source, **options):
    warnings = []
    tables = []

    # Load datapackage
    try:
        package = Package(source, **options)
    except exceptions.DataPackageException as error:
        warnings.append(
            'Data Package "%s" has a loading error "%s"' %
            (source, error))

    # Extract datapackage tables
    if not warnings:
        for resource in package.resources:
            if resource.tabular:
                tables.append({
                    'source': resource.source if not resource.inline else 'inline',
                    'stream': Stream(partial(_iter_resource_rows, resource), headers=1),
                    'schema': resource.schema,
                    'extra': {
                        'datapackage': str(source),
                    },
                })

    # Extrace datapackage errors
    if not warnings:
        for error in package.errors:
            warnings.append(
                'Data Package "%s" has a validation error "%s"' %
                (source, error))

    return warnings, tables


# Internal

def _iter_resource_rows(resource):
    for index, row in enumerate(resource.iter(cast=False)):
        if not index:
            yield resource.headers
        yield row



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import functools
from .spec import spec


@functools.total_ordering
class Error(object):
    """Describes a validation check error.

    Args:
        code (str): The error code. Must be one in the spec.
        cell (dict, optional): The cell where the error occurred.
        row_number (int, optional): The row number where the error occurs.
        message (str, optional The error message. Defaults to the message from
            the Data Quality Spec.
        message_substitutions (dict, optional Dictionary with substitutions to
            be used when generating the error message and description.

    Raises:
        KeyError: Raised if the error code isn't known.
    """

    def __init__(
        self,
        code,
        cell=None,
        row_number=None,
        message=None,
        message_substitutions=None
    ):
        self._spec = spec['errors'].get(code)
        default_message = None
        if self._spec:
            default_message = self._spec['message']

        self._code = code
        self._cell = cell or {}
        self._row = None
        self._row_number = row_number or self._cell.get('row-number')
        self._message = message or default_message
        self._message_substitutions = message_substitutions or {}

    def __iter__(self):
        for key, value in self._to_dict().items():
            yield (key, value)

    @property
    def cell(self):
        return self._cell

    @property
    def code(self):
        return self._code

    @property
    def row_number(self):
        return self._row_number

    @property
    def column_number(self):
        return self._cell.get('column-number')

    @property
    def message(self):
        if self._message:
            return self._message.format(
                row_number=self.row_number,
                column_number=self.column_number,
                **self._message_substitutions
            )

    @property
    def description(self):
        # TODO: Add string substitutions
        if self._spec:
            return self._spec['description']

    @property
    def row(self):
        return self._row

    @row.setter
    def row(self, value):
        self._row = value

    def __eq__(self, other):
        my_row = self.row_number or 0
        my_col = self.column_number or 0
        other_row = other.row_number or 0
        other_col = other.column_number or 0

        return (my_row == other_row) and (my_col == other_col)

    def __lt__(self, other):
        my_row = self.row_number or 0
        my_col = self.column_number or 0
        other_row = other.row_number or 0
        other_col = other.column_number or 0

        return (my_row < other_row) or (my_row == other_row and my_col < other_col)

    def _to_dict(self):
        return {
            'code': self.code,
            'row-number': self.row_number,
            'column-number': self.column_number,
            'message': self.message,
            'message-data': {k: v.strip('"') if isinstance(v, str) else v for k, v
                             in self._message_substitutions.items()},
        }
